/* HEADER
If you use this code dont forget to reference us :) BibTeX: http://www.cs.technion.ac.il/~rani/el-yaniv_bib.html#BegleiterEY04

This code is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This code is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License (<a href="http://www.gnu.org/copyleft/gpl.html">GPL</a>) for more details.*/

package weka.classifiers.compression.vmm.algs;

import java.io.Serializable;
import java.io.UnsupportedEncodingException;

import weka.classifiers.compression.vmm.pred.VMMPredictor;
import weka.classifiers.compression.vmm.pred.VMMNotTrainedException;

import weka.classifiers.compression.vmm.algs.lzms.*;
import gnu.trove.*;

/**
 * <b>LZms Predictor</b> Usage example: LZmsPredictor lzms = new
 * LZmsPredictor(); lzms.init(256, 2, 0); lzms.learn("abracadabra");
 * System.out.println("logeval : " + lzms.logEval("cadabra"));
 * System.out.println("P(c|abra) : " + lzms.predict('c', "abra"));
 * 
 * Using Moti Nisenson's code.
 * <p>
 * Copyright: Copyright (c) 2004
 * </p>
 * 
 * @author <a href="http://www.cs.technion.ac.il/~ronbeg">Ron Begleiter</a>
 * @version 1.0
 */

public class LZmsPredictor implements VMMPredictor, Serializable {

	private static final long serialVersionUID = -8381652570701322374L;

	private LZmsTree lzms;

	private int abSize;
	private int mParam;
	private int sParam;

	public LZmsPredictor() {
	}

	public void init(int abSize, int mParam, int sParam) {
		this.abSize = abSize;
		this.mParam = mParam;
		this.sParam = sParam;
		lzms = new LZmsTree(mParam, sParam, abSize);
	}

	public void learn(String trainingSequence) {
		String newString = "";
		try {
			newString = new String(trainingSequence.getBytes("ISO-8859-1"));
		} catch (UnsupportedEncodingException e1) {
			// TODO Auto-generated catch block
			e1.printStackTrace();
		}
		TIntArrayList symList = new TIntArrayList();

		for (int i = 0; i < newString.length(); ++i) {
			symList.add((int) newString.charAt(i));
		}

		lzms.learnSequence(symList.toNativeArray());
	}

	public double predict(int symbol, String context) {
		try {

			// using the following relation: P(a | X) = P(Xa)/P(X)
			StringBuffer seqWithSym = new StringBuffer();
			seqWithSym.append(context).append((char) symbol);
			StringBuffer seqNoSym = new StringBuffer();
			seqNoSym.append(context);

			return Math.pow(2.0,
					-(logEval(seqWithSym.toString()) - logEval(seqNoSym
							.toString())));
		} catch (NullPointerException npe) {
			if (lzms == null) {
				throw new VMMNotTrainedException();
			} else {
				throw npe;
			}
		}

	}

	public double logEval(String testSequence) {
		try {
			String newString = "";
			try {
				newString = new String(testSequence.getBytes("ISO-8859-1"));
			} catch (UnsupportedEncodingException e1) {
				// TODO Auto-generated catch block
				e1.printStackTrace();
			}
			TIntArrayList symList = new TIntArrayList();
			for (int i = 0; i < newString.length(); ++i) {
				symList.add((int) newString.charAt(i));
			}
			return lzms.calcLogLikelihood(symList.toNativeArray());
		} catch (NullPointerException npe) {
			if (lzms == null) {
				throw new VMMNotTrainedException();
			} else {
				throw npe;
			}
		}

	}

	public double logEval(String testSequence, String initialContext) {
		logEval(initialContext);
		return logEval(testSequence);
	}

}
